/*  This file is part of the OpenLB library
 *
 *  Copyright (C) 2006, 2007 Jonas Latt
 *  Address: Rue General Dufour 24,  1211 Geneva 4, Switzerland 
 *  E-mail: jonas.latt@gmail.com
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public 
 *  License along with this program; if not, write to the Free 
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
*/

/** \file
 * Helper functions for the implementation of LB dynamics. This file is all
 * about efficiency. The generic template code is specialized for commonly
 * used Lattices, so that a maximum performance can be taken out of each
 * case.
 */
#ifndef LB_HELPERS_H
#define LB_HELPERS_H

#include "latticeDescriptors.h"
#include "cell.h"
#include "util.h"


namespace olb {

/// All helper functions are inside this structure
template<typename T, template<typename U> class Lattice>
struct lbHelpers {
    /// Computation of equilibrium distribution
    static T equilibrium( int iPop, T rho, const T u[Lattice<T>::d],
                          const T uSqr )
    {
        T c_u = T();
        for (int iD=0; iD < Lattice<T>::d; ++iD) {
           c_u += Lattice<T>::c[iPop][iD]*u[iD];
        }
        return rho * Lattice<T>::t[iPop] * (
               (T)1 + Lattice<T>::invCs2 * c_u +
               Lattice<T>::invCs2 * Lattice<T>::invCs2/(T)2 * c_u*c_u -
               Lattice<T>::invCs2/(T)2 * uSqr
           ) - Lattice<T>::t[iPop];
    }

    static T incEquilibrium( int iPop, const T j[Lattice<T>::d],
                             const T jSqr, const T pressure )
    {
        T c_j = T();
        for (int iD=0; iD < Lattice<T>::d; ++iD) {
           c_j += Lattice<T>::c[iPop][iD]*j[iD];
        }
        T rho = (T)1 + pressure*Lattice<T>::invCs2;
        return Lattice<T>::t[iPop] * ( rho +
               Lattice<T>::invCs2 * c_j +
               Lattice<T>::invCs2 * Lattice<T>::invCs2/(T)2 * c_j*c_j -
               Lattice<T>::invCs2/(T)2 * jSqr
           ) - Lattice<T>::t[iPop];
    }

    static void computeFneq (
            Cell<T,Lattice> const& cell,
            T fNeq[Lattice<T>::q], T rho, const T u[Lattice<T>::d] )
    {
        const T uSqr = util::normSqr<T,Lattice<T>::d>(u);
        for (int iPop=0; iPop < Lattice<T>::q; ++iPop) {
            fNeq[iPop] = cell[iPop] - equilibrium(iPop, rho, u, uSqr);
        }
    }


    /// BGK collision step
    static T bgkCollision( Cell<T,Lattice>& cell,
                           T rho, const T u[Lattice<T>::d],
                           T omega)
    {
        const T uSqr = util::normSqr<T,Lattice<T>::d>(u);
        for (int iPop=0; iPop < Lattice<T>::q; ++iPop) {
            cell[iPop] *= (T)1-omega;
            cell[iPop] += omega * lbHelpers<T,Lattice>::equilibrium (
                              iPop, rho, u, uSqr );
        }
        return uSqr;
    }

    /// Incompressible BGK collision step
    static T incBgkCollision( Cell<T,Lattice>& cell,
                              T pressure,
                              const T j[Lattice<T>::d],
                              T omega)
    {
        const T jSqr = util::normSqr<T,Lattice<T>::d>(j);
        for (int iPop=0; iPop < Lattice<T>::q; ++iPop) {
            cell[iPop] *= (T)1-omega;
            cell[iPop] += omega * lbHelpers<T,Lattice>::incEquilibrium (
                              iPop, j, jSqr, pressure );
        }
        return jSqr;
    }

    /// BGK collision step with density correction
    static T constRhoBgkCollision( Cell<T,Lattice>& cell,
                                   T rho, const T u[Lattice<T>::d],
                                   T ratioRho, T omega)
    {
        const T uSqr = util::normSqr<T,Lattice<T>::d>(u);
        for (int iPop=0; iPop < Lattice<T>::q; ++iPop) {
            T feq = lbHelpers<T,Lattice>::equilibrium(iPop, rho, u, uSqr );
            cell[iPop] =
              ratioRho*(feq+Lattice<T>::t[iPop])-Lattice<T>::t[iPop] +
              ((T)1-omega)*(cell[iPop]-feq);
        }
        return uSqr;
    }


    /// Add a force term after BGK collision
    static void addExternalForce( Cell<T,Lattice>& cell,
                                  const T u[Lattice<T>::d],
                                  T omega )
    {
        static const int forceBeginsAt
            = Lattice<T>::ExternalField::forceBeginsAt;
        T* force = cell.getExternal(forceBeginsAt);
        for (int iPop=0; iPop < Lattice<T>::q; ++iPop) {
            T c_u = T();
            for (int iD=0; iD < Lattice<T>::d; ++iD) {
               c_u += Lattice<T>::c[iPop][iD]*u[iD];
            }
            c_u *= Lattice<T>::invCs2 * Lattice<T>::invCs2;
            T forceTerm = T();
            for (int iD=0; iD < Lattice<T>::d; ++iD) {
                forceTerm +=
                    (   (Lattice<T>::c[iPop][iD]-u[iD]) * Lattice<T>::invCs2
                      + c_u * Lattice<T>::c[iPop][iD]
                    )
                    * force[iD];
            }
            forceTerm *= Lattice<T>::t[iPop];
            forceTerm *= 1-omega/(T)2;
            cell[iPop] += forceTerm;
        }
    }


    /// Computation of density
    static T computeRho (
        Cell<T,Lattice> const& cell)
    {
        T rho = T();
        for (int iPop=0; iPop < Lattice<T>::q; ++iPop) {
            rho += cell[iPop];
        }
        rho += (T)1;
        return rho;
    }

    /// Computation of momentum
    static void computeJ (
        Cell<T,Lattice> const& cell, T j[Lattice<T>::d] )
    {
        for (int iD=0; iD < Lattice<T>::d; ++iD) {
            j[iD] = T();
        }
        for (int iPop=0; iPop < Lattice<T>::q; ++iPop) {
            for (int iD=0; iD < Lattice<T>::d; ++iD) {
                j[iD] += cell[iPop]*Lattice<T>::c[iPop][iD];
            }
        }
    }

    /// Computation of hydrodynamic variables
    static void computeRhoU (
        Cell<T,Lattice> const& cell, T& rho, T u[Lattice<T>::d] )
    {
        rho = T();
        for (int iD=0; iD < Lattice<T>::d; ++iD) {
            u[iD] = T();
        }
        for (int iPop=0; iPop < Lattice<T>::q; ++iPop) {
            rho += cell[iPop];
            for (int iD=0; iD < Lattice<T>::d; ++iD) {
                u[iD] += cell[iPop]*Lattice<T>::c[iPop][iD];
            }
        }
        rho += (T)1;
        for (int iD=0; iD < Lattice<T>::d; ++iD) {
            u[iD] /= rho;
        }
    }

    /// Computation of stress tensor
    static void computeStress (
        Cell<T,Lattice> const& cell, T rho, const T u[Lattice<T>::d],
        T pi[util::TensorVal<Lattice<T> >::n] )
    {
        int iPi = 0;
        for (int iAlpha=0; iAlpha < Lattice<T>::d; ++iAlpha) {
            for (int iBeta=iAlpha; iBeta < Lattice<T>::d; ++iBeta) {
                pi[iPi] = T();
                for (int iPop=0; iPop < Lattice<T>::q; ++iPop) {
                    pi[iPi] += Lattice<T>::c[iPop][iAlpha]*
                               Lattice<T>::c[iPop][iBeta] * cell[iPop];
                }
                // stripe off equilibrium contribution
                pi[iPi] -= rho*u[iAlpha]*u[iBeta];
                if (iAlpha==iBeta) {
                    pi[iPi] -= 1./Lattice<T>::invCs2*(rho-(T)1);
                }
                ++iPi;
            }
        }
    }
    
    /// Computation of all hydrodynamic variables
    static void computeAllMomenta (
        Cell<T,Lattice> const& cell, T& rho, T u[Lattice<T>::d],
        T pi[util::TensorVal<Lattice<T> >::n] )
    {
        computeRhoU(cell, rho, u);
        computeStress(cell, rho, u, pi);
    }

    /// Swap ("bounce-back") values of a cell (2D), and apply streaming step
    static void swapAndStream2D(Cell<T,Lattice> **grid, int iX, int iY)
    {
        const int half = Lattice<T>::q/2;
        for (int iPop=1; iPop<=half; ++iPop) {
            int nextX = iX + Lattice<T>::c[iPop][0];
            int nextY = iY + Lattice<T>::c[iPop][1];
            T fTmp                   = grid[iX][iY][iPop];
            grid[iX][iY][iPop]       = grid[iX][iY][iPop+half];
            grid[iX][iY][iPop+half]  = grid[nextX][nextY][iPop];
            grid[nextX][nextY][iPop] = fTmp;
         }
    }

    /// Swap ("bounce-back") values of a cell (3D), and apply streaming step
    static void swapAndStream3D(Cell<T,Lattice> ***grid,
                                int iX, int iY, int iZ)
    {
        const int half = Lattice<T>::q/2;
        for (int iPop=1; iPop<=half; ++iPop) {
            int nextX = iX + Lattice<T>::c[iPop][0];
            int nextY = iY + Lattice<T>::c[iPop][1];
            int nextZ = iZ + Lattice<T>::c[iPop][2];
            T fTmp                          = grid[iX][iY][iZ][iPop];
            grid[iX][iY][iZ][iPop]          = grid[iX][iY][iZ][iPop+half];
            grid[iX][iY][iZ][iPop+half]     = grid[nextX][nextY][nextZ][iPop];
            grid[nextX][nextY][nextZ][iPop] = fTmp;
         }
    }

    static void modifyVelocity(Cell<T,Lattice>& cell,
                               const T newU[Lattice<T>::d])
    {
        T rho, oldU[Lattice<T>::d];
        computeRhoU(cell, rho, oldU);
        const T oldUSqr = util::normSqr<T,Lattice<T>::d>(oldU);
        const T newUSqr = util::normSqr<T,Lattice<T>::d>(newU);
        for (int iPop=0; iPop<Lattice<T>::q; ++iPop) {
            cell[iPop] = cell[iPop]
                             - equilibrium(iPop, rho, oldU, oldUSqr)
                             + equilibrium(iPop, rho, newU, newUSqr);
        }
    }

};  // struct lbHelpers

/// Helper functions for dynamics that access external field
template<typename T, template<typename U> class Lattice>
struct externalFieldHelpers {
    /// Add a force term after BGK collision
    static void addExternalForce( Cell<T,Lattice>& cell,
                                  const T u[Lattice<T>::d],
                                  T omega,
                                  T amplitude=(T)1)
    {
        static const int forceBeginsAt
            = Lattice<T>::ExternalField::forceBeginsAt;
        T* force = cell.getExternal(forceBeginsAt);
        for (int iPop=0; iPop < Lattice<T>::q; ++iPop) {
            T c_u = T();
            for (int iD=0; iD < Lattice<T>::d; ++iD) {
               c_u += Lattice<T>::c[iPop][iD]*u[iD];
            }
            c_u *= Lattice<T>::invCs2 * Lattice<T>::invCs2;
            T forceTerm = T();
            for (int iD=0; iD < Lattice<T>::d; ++iD) {
                forceTerm +=
                    (   (Lattice<T>::c[iPop][iD]-u[iD]) * Lattice<T>::invCs2
                      + c_u * Lattice<T>::c[iPop][iD]
                    )
                    * force[iD];
            }
            forceTerm *= Lattice<T>::t[iPop];
            forceTerm *= 1-omega/(T)2;
            forceTerm *= amplitude;
            cell[iPop] += forceTerm;
        }
    }
};  // struct externalFieldHelpers


/// All boundary helper functions are inside this structure
template<typename T, template<typename U> class Lattice,
         int direction, int orientation>
struct BoundaryHelpers {
    static void computeStress (
         Cell<T,Lattice> const& cell,
         T rho, const T u[Lattice<T>::d],
         T pi[util::TensorVal<Lattice<T> >::n] )
    {
        typedef Lattice<T> L;
        const T uSqr = util::normSqr<T,L::d>(u);

        std::vector<int> const& onWallIndices = util::subIndex<L, direction, 0>();
        std::vector<int> const& normalIndices = util::subIndex<L, direction, orientation>();

        T fNeq[Lattice<T>::q];
        for (unsigned fIndex=0; fIndex<onWallIndices.size(); ++fIndex) {
            int iPop = onWallIndices[fIndex];
            fNeq[iPop] =
                cell[iPop] -
                lbHelpers<T,Lattice>::equilibrium(iPop, rho, u, uSqr);
        }
        for (unsigned fIndex=0; fIndex<normalIndices.size(); ++fIndex) {
            int iPop = normalIndices[fIndex];
            if (iPop == 0) {
                fNeq[iPop] = T();  // fNeq[0] will not be used anyway
            }
            else {
                fNeq[iPop] =
                    cell[iPop] -
                    lbHelpers<T,Lattice>::equilibrium(iPop, rho, u, uSqr);
            }
        }

        int iPi = 0;
        for (int iAlpha=0; iAlpha<L::d; ++iAlpha) {
            for (int iBeta=iAlpha; iBeta<L::d; ++iBeta) {
                pi[iPi] = T();
                for (unsigned fIndex=0; fIndex<onWallIndices.size(); ++fIndex)
                {
                    const int iPop = onWallIndices[fIndex];
                    pi[iPi] +=
                        L::c[iPop][iAlpha]*L::c[iPop][iBeta]*fNeq[iPop];
                }
                for (unsigned fIndex=0; fIndex<normalIndices.size(); ++fIndex)
                {
                    const int iPop = normalIndices[fIndex];
                    pi[iPi] += (T)2 * L::c[iPop][iAlpha]*L::c[iPop][iBeta]*
                               fNeq[iPop];
                }
                ++iPi;
            }
        }
    }

};  // struct boundaryHelpers

}  // namespace olb

// The specialized code is directly included. That is because we never want
// it to be precompiled so that in both the precompiled and the
// "include-everything" version, the compiler can apply all the
// optimizations it wants.
#include "lbHelpers2D.h"
#include "lbHelpers3D.h"

#endif
